@@ -45,15 +45,13 @@
 #include <stdlib.h>
 #include <sys/errno.h>
 #include <sys/types.h>
-#include <netinet/in.h>
+#include <stdint.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
 #include <arpa/inet.h>
 #include <net/if.h>
 #include <netinet/in.h>
 #include <netdb.h>
-#include <unistd.h>
-#include <fcntl.h>
 
 #include "rpc_subs.h"
 #include "replicated.h"
@@ -62,7 +60,10 @@
 #ifndef MAX_ERR_BUF
 #define MAX_ERR_BUF		512
 #endif
+
 #define MAX_IFC_BUF		2048
+static int volatile ifc_buf_len = MAX_IFC_BUF;
+static int volatile ifc_last_len = 0;
 
 #define MASK_A  0x7F000000
 #define MASK_B  0xBFFF0000
@@ -74,12 +75,26 @@
 #define max(x, y)	(x >= y ? x : y)
 #define mmax(x, y, z)	(max(x, y) == x ? max(x, z) : max(y, z))
 
+unsigned int ipv6_mask_cmp(uint32_t *host, uint32_t *iface, uint32_t *mask)
+{
+	unsigned int ret = 1;
+	unsigned int i;
+
+	for (i = 0; i < 4; i++) {
+		if ((host[i] & mask[i]) != (iface[i] & mask[i])) {
+			ret = 0;
+			break;
+		}
+	}
+	return ret;
+}
+
 void seed_random(void)
 {
 	int fd;
 	unsigned int seed;
 
-	fd = open("/dev/random", O_RDONLY);
+	fd = open_fd("/dev/urandom", O_RDONLY);
 	if (fd < 0) {
 		srandom(time(NULL));
 		return;
@@ -95,39 +110,98 @@ void seed_random(void)
 	return;
 }
 
-static unsigned int get_proximity(const char *host_addr, int addr_len)
+static int alloc_ifreq(struct ifconf *ifc, int sock)
 {
-	struct sockaddr_in *msk_addr, *if_addr;
+	int ret, lastlen = ifc_last_len, len = ifc_buf_len;
+	char err_buf[MAX_ERR_BUF], *buf;
+
+	while (1) {
+		buf = malloc(len);
+		if (!buf) {
+			char *estr = strerror_r(errno, err_buf, MAX_ERR_BUF);
+			logerr("malloc: %s", estr);
+			return 0;
+		}
+
+		ifc->ifc_len = len;
+		ifc->ifc_req = (struct ifreq *) buf;
+
+		ret = ioctl(sock, SIOCGIFCONF, ifc);
+		if (ret == -1) {
+			char *estr = strerror_r(errno, err_buf, MAX_ERR_BUF);
+			logerr("ioctl: %s", estr);
+			free(buf);
+			return 0;
+		}
+
+		if (ifc->ifc_len <= lastlen)
+			break;
+
+		lastlen = ifc->ifc_len;
+		len += MAX_IFC_BUF;
+		free(buf);
+	}
+
+	if (lastlen != ifc_last_len) {
+		ifc_last_len = lastlen;
+		ifc_buf_len = len;
+	}
+
+	return 1;
+}
+
+static unsigned int get_proximity(struct sockaddr *host_addr)
+{
+	struct sockaddr_in *addr, *msk_addr, *if_addr;
+	struct sockaddr_in6 *addr6, *msk6_addr, *if6_addr;
 	struct in_addr *hst_addr;
-	char tmp[20], buf[MAX_ERR_BUF], *ptr;
+	struct in6_addr *hst6_addr;
+	int addr_len;
+	char buf[MAX_ERR_BUF], *ptr;
 	struct ifconf ifc;
 	struct ifreq *ifr, nmptr;
-	int sock, cl_flags, ret, i;
-	uint32_t mask, ha, ia;
+	int sock, ret, i;
+	uint32_t mask, ha, ia, *mask6, *ha6, *ia6;
 
-	memcpy(tmp, host_addr, addr_len);
-	hst_addr = (struct in_addr *) tmp;
+	addr = NULL;
+	addr6 = NULL;
+	hst_addr = NULL;
+	hst6_addr = NULL;
+	mask6 = NULL;
+	ha6 = NULL;
+	ia6 = NULL;
 
-	ha = ntohl((uint32_t) hst_addr->s_addr);
+	switch (host_addr->sa_family) {
+	case AF_INET:
+		addr = (struct sockaddr_in *) host_addr;
+		hst_addr = (struct in_addr *) &addr->sin_addr;
+		ha = ntohl((uint32_t) hst_addr->s_addr);
+		addr_len = sizeof(hst_addr);
+		break;
 
-	sock = socket(AF_INET, SOCK_DGRAM, 0);
+	case AF_INET6:
+#ifndef INET6
+		return PROXIMITY_UNSUPPORTED;
+#else
+		addr6 = (struct sockaddr_in6 *) host_addr;
+		hst6_addr = (struct in6_addr *) &addr6->sin6_addr;
+		ha6 = &hst6_addr->s6_addr32[0];
+		addr_len = sizeof(hst6_addr);
+		break;
+#endif
+
+	default:
+		return PROXIMITY_ERROR;
+	}
+
+	sock = open_sock(AF_INET, SOCK_DGRAM, 0);
 	if (sock < 0) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 		logerr("socket creation failed: %s", estr);
 		return PROXIMITY_ERROR;
 	}
 
-	if ((cl_flags = fcntl(sock, F_GETFD, 0)) != -1) {
-		cl_flags |= FD_CLOEXEC;
-		fcntl(sock, F_SETFD, cl_flags);
-	}
-
-	ifc.ifc_len = sizeof(buf);
-	ifc.ifc_req = (struct ifreq *) buf;
-	ret = ioctl(sock, SIOCGIFCONF, &ifc);
-	if (ret == -1) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr("ioctl: %s", estr);
+	if (!alloc_ifreq(&ifc, sock)) {
 		close(sock);
 		return PROXIMITY_ERROR;
 	}
@@ -138,19 +212,36 @@ static unsigned int get_proximity(const char *host_addr, int addr_len)
 	i = 0;
 	ptr = (char *) &ifc.ifc_buf[0];
 
-	while (ptr < buf + ifc.ifc_len) {
+	while (ptr < (char *) ifc.ifc_req + ifc.ifc_len) {
 		ifr = (struct ifreq *) ptr;
 
 		switch (ifr->ifr_addr.sa_family) {
 		case AF_INET:
+			if (host_addr->sa_family == AF_INET6)
+				break;
 			if_addr = (struct sockaddr_in *) &ifr->ifr_addr;
 			ret = memcmp(&if_addr->sin_addr, hst_addr, addr_len);
 			if (!ret) {
 				close(sock);
+				free(ifc.ifc_req);
 				return PROXIMITY_LOCAL;
 			}
 			break;
 
+		case AF_INET6:
+#ifdef INET6
+			if (host_addr->sa_family == AF_INET)
+				break;
+
+			if6_addr = (struct sockaddr_in6 *) &ifr->ifr_addr;
+			ret = memcmp(&if6_addr->sin6_addr, hst6_addr, addr_len);
+			if (!ret) {
+				close(sock);
+				free(ifc.ifc_req);
+				return PROXIMITY_LOCAL;
+			}
+#endif
+
 		default:
 			break;
 		}
@@ -162,30 +253,34 @@ static unsigned int get_proximity(const char *host_addr, int addr_len)
 	i = 0;
 	ptr = (char *) &ifc.ifc_buf[0];
 
-	while (ptr < buf + ifc.ifc_len) {
+	while (ptr < (char *) ifc.ifc_req + ifc.ifc_len) {
 		ifr = (struct ifreq *) ptr;
 
+		nmptr = *ifr;
+		ret = ioctl(sock, SIOCGIFNETMASK, &nmptr);
+		if (ret == -1) {
+			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+			logerr("ioctl: %s", estr);
+			close(sock);
+			free(ifc.ifc_req);
+			return PROXIMITY_ERROR;
+		}
+
 		switch (ifr->ifr_addr.sa_family) {
 		case AF_INET:
+			if (host_addr->sa_family == AF_INET6)
+				break;
 			if_addr = (struct sockaddr_in *) &ifr->ifr_addr;
 			ia =  ntohl((uint32_t) if_addr->sin_addr.s_addr);
 
 			/* Is the address within a localiy attached subnet */
 
-			nmptr = *ifr;
-			ret = ioctl(sock, SIOCGIFNETMASK, &nmptr);
-			if (ret == -1) {
-				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				logerr("ioctl: %s", estr);
-				close(sock);
-				return PROXIMITY_ERROR;
-			}
-
 			msk_addr = (struct sockaddr_in *) &nmptr.ifr_netmask;
 			mask = ntohl((uint32_t) msk_addr->sin_addr.s_addr);
 
 			if ((ia & mask) == (ha & mask)) {
 				close(sock);
+				free(ifc.ifc_req);
 				return PROXIMITY_SUBNET;
 			}
 
@@ -208,10 +303,34 @@ static unsigned int get_proximity(const char *host_addr, int addr_len)
 
 			if ((ia & mask) == (ha & mask)) {
 				close(sock);
+				free(ifc.ifc_req);
 				return PROXIMITY_NET;
 			}
 			break;
 
+		case AF_INET6:
+#ifdef INET6
+			if (host_addr->sa_family == AF_INET)
+				break;
+
+			if6_addr = (struct sockaddr_in6 *) &ifr->ifr_addr;
+			ia6 = &if6_addr->sin6_addr.s6_addr32[0];
+
+			/* Is the address within the network of the interface */
+
+			msk6_addr = (struct sockaddr_in6 *) &nmptr.ifr_netmask;
+			mask6 = &msk6_addr->sin6_addr.s6_addr32[0];
+
+			if (ipv6_mask_cmp(ha6, ia6, mask6)) {
+				close(sock);
+				free(ifc.ifc_req);
+				return PROXIMITY_SUBNET;
+			}
+
+			/* How do we define "local network" in ipv6? */
+#endif
+			break;
+
 		default:
 			break;
 		}
@@ -221,14 +340,19 @@ static unsigned int get_proximity(const char *host_addr, int addr_len)
 	}
 
 	close(sock);
+	free(ifc.ifc_req);
 
 	return PROXIMITY_OTHER;
 }
 
-static struct host *new_host(const char *name, const char *addr, unsigned int proximity, unsigned int weight)
+static struct host *new_host(const char *name,
+			     struct sockaddr *addr, size_t addr_len,
+			     unsigned int proximity, unsigned int weight,
+			     unsigned int options)
 {
 	struct host *new;
-	char *tmp1, *tmp2;
+	struct sockaddr *tmp2;
+	char *tmp1;
 
 	if (!name || !addr)
 		return NULL;
@@ -237,11 +361,12 @@ static struct host *new_host(const char *name, const char *addr, unsigned int pr
 	if (!tmp1)
 		return NULL;
 
-	tmp2 = strdup(addr);
+	tmp2 = malloc(addr_len);
 	if (!tmp2) {
 		free(tmp1);
 		return NULL;
 	}
+	memcpy(tmp2, addr, addr_len);
 
 	new = malloc(sizeof(struct host));
 	if (!new) {
@@ -253,9 +378,11 @@ static struct host *new_host(const char *name, const char *addr, unsigned int pr
 	memset(new, 0, sizeof(struct host));
 
 	new->name = tmp1;
+	new->addr_len = addr_len;
 	new->addr = tmp2;
 	new->proximity = proximity;
 	new->weight = weight;
+	new->options = options;
 
 	return new;
 }
@@ -264,7 +391,7 @@ static int add_host(struct host **list, struct host *host)
 {
 	struct host *this, *last;
 
-	if (!list) {
+	if (!*list) {
 		*list = host;
 		return 1;
 	}
@@ -390,9 +517,13 @@ static unsigned short get_port_option(const char *options)
 static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 			 struct conn_info *pm_info, struct conn_info *rpc_info,
 			 const char *proto, unsigned int version,
-			 const char *options, unsigned int random_selection)
+			 const char *options)
 {
 	char *have_port_opt = options ? strstr(options, "port=") : NULL;
+	unsigned int random_selection = host->options & MOUNT_FLAG_RANDOM_SELECT;
+	unsigned int use_weight_only = host->options & MOUNT_FLAG_USE_WEIGHT_ONLY;
+	socklen_t len = INET6_ADDRSTRLEN;
+	char buf[len + 1];
 	struct pmap parms;
 	struct timeval start, end;
 	struct timezone tz;
@@ -400,6 +531,15 @@ static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 	double taken = 0;
 	int status, count = 0;
 
+	if (host->addr)
+		debug(logopt, "called with host %s(%s) proto %s version 0x%x",
+		      host->name, get_addr_string(host->addr, buf, len),
+		      proto, version);
+	else
+		debug(logopt,
+		      "called for host %s proto %s version 0x%x",
+		      host->name, proto, version);
+
 	memset(&parms, 0, sizeof(struct pmap));
 
 	parms.pm_prog = NFS_PROGRAM;
@@ -424,11 +564,17 @@ static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 		status = rpc_ping_proto(rpc_info);
 		gettimeofday(&end, &tz);
 		if (status) {
-			if (random_selection)
+			double reply;
+			if (random_selection) {
 				/* Random value between 0 and 1 */
-				taken += ((float) random())/((float) RAND_MAX+1);
-			else
-				taken += elapsed(start, end);;
+				reply = ((float) random())/((float) RAND_MAX+1);
+				debug(logopt,
+				      "nfs v4 random selection time: %f", reply);
+			} else {
+				reply = elapsed(start, end);
+				debug(logopt, "nfs v4 rpc ping time: %f", reply);
+			}
+			taken += reply;
 			count++;
 			supported = NFS4_SUPPORTED;
 		}
@@ -437,7 +583,8 @@ static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 v3_ver:
 	if (!have_port_opt) {
 		status = rpc_portmap_getclient(pm_info,
-				 host->name, proto, RPC_CLOSE_DEFAULT);
+				host->name, host->addr, host->addr_len,
+				proto, RPC_CLOSE_DEFAULT);
 		if (!status)
 			goto done_ver;
 	}
@@ -465,11 +612,17 @@ v3_ver:
 		status = rpc_ping_proto(rpc_info);
 		gettimeofday(&end, &tz);
 		if (status) {
-			if (random_selection)
+			double reply;
+			if (random_selection) {
 				/* Random value between 0 and 1 */
-				taken += ((float) random())/((float) RAND_MAX+1);
-			else
-				taken += elapsed(start, end);;
+				reply = ((float) random())/((float) RAND_MAX+1);
+				debug(logopt,
+				      "nfs v3 random selection time: %f", reply);
+			} else {
+				reply = elapsed(start, end);
+				debug(logopt, "nfs v3 rpc ping time: %f", reply);
+			}
+			taken += reply;
 			count++;
 			supported |= NFS3_SUPPORTED;
 		}
@@ -499,11 +652,17 @@ v2_ver:
 		status = rpc_ping_proto(rpc_info);
 		gettimeofday(&end, &tz);
 		if (status) {
-			if (random_selection)
+			double reply;
+			if (random_selection) {
 				/* Random value between 0 and 1 */
-				taken += ((float) random())/((float) RAND_MAX+1);
-			else
-				taken += elapsed(start, end);;
+				reply = ((float) random())/((float) RAND_MAX+1);
+				debug(logopt,
+				      "nfs v2 random selection time: %f", reply);
+			} else {
+				reply = elapsed(start, end);;
+				debug(logopt, "nfs v2 rpc ping time: %f", reply);
+			}
+			taken += reply;
 			count++;
 			supported |= NFS2_SUPPORTED;
 		}
@@ -523,19 +682,24 @@ done_ver:
 		 * Average response time to 7 significant places as
 		 * integral type.
 		 */
-		host->cost = (unsigned long) ((taken * 1000000) / count);
+		if (use_weight_only)
+			host->cost = 1;
+		else
+			host->cost = (unsigned long) ((taken * 1000000) / count);
 
 		/* Allow for user bias */
 		if (host->weight)
 			host->cost *= (host->weight + 1);
+
+		debug(logopt, "host %s cost %ld weight %d",
+		      host->name, host->cost, host->weight);
 	}
 
 	return supported;
 }
 
 static int get_vers_and_cost(unsigned logopt, struct host *host,
-			     unsigned int version, const char *options,
-			     unsigned int random_selection)
+			     unsigned int version, const char *options)
 {
 	struct conn_info pm_info, rpc_info;
 	time_t timeout = RPC_TIMEOUT;
@@ -547,10 +711,12 @@ static int get_vers_and_cost(unsigned logopt, struct host *host,
 
 	if (host->proximity == PROXIMITY_NET)
 		timeout = RPC_TIMEOUT * 2;
-	else if (host->proximity == PROXIMITY_NET)
+	else if (host->proximity == PROXIMITY_OTHER)
 		timeout = RPC_TIMEOUT * 8;
 
 	rpc_info.host = host->name;
+	rpc_info.addr = host->addr;
+	rpc_info.addr_len = host->addr_len;
 	rpc_info.program = NFS_PROGRAM;
 	rpc_info.timeout.tv_sec = timeout;
 	rpc_info.close_option = RPC_CLOSE_DEFAULT;
@@ -560,8 +726,7 @@ static int get_vers_and_cost(unsigned logopt, struct host *host,
 
 	if (version & UDP_REQUESTED) {
 		supported = get_nfs_info(logopt, host,
-					&pm_info, &rpc_info, "udp", vers,
-					options, random_selection);
+				   &pm_info, &rpc_info, "udp", vers, options);
 		if (supported) {
 			ret = 1;
 			host->version |= (supported << 8);
@@ -570,8 +735,7 @@ static int get_vers_and_cost(unsigned logopt, struct host *host,
 
 	if (version & TCP_REQUESTED) {
 		supported = get_nfs_info(logopt, host,
-					 &pm_info, &rpc_info, "tcp", vers,
-					 options, random_selection);
+				   &pm_info, &rpc_info, "tcp", vers, options);
 		if (supported) {
 			ret = 1;
 			host->version |= supported;
@@ -582,10 +746,13 @@ static int get_vers_and_cost(unsigned logopt, struct host *host,
 }
 
 static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
-				      unsigned int version, const char *options,
-				      unsigned int random_selection)
+				      unsigned int version, const char *options)
 {
 	char *have_port_opt = options ? strstr(options, "port=") : NULL;
+	unsigned int random_selection = host->options & MOUNT_FLAG_RANDOM_SELECT;
+	unsigned int use_weight_only = host->options & MOUNT_FLAG_USE_WEIGHT_ONLY;
+	socklen_t len = INET6_ADDRSTRLEN;
+	char buf[len + 1];
 	struct conn_info pm_info, rpc_info;
 	struct pmap parms;
 	const char *proto;
@@ -596,16 +763,26 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 	time_t timeout = RPC_TIMEOUT;
 	int status;
 
+	if (host->addr)
+		debug(logopt, "called with host %s(%s) version 0x%x",
+			host->name, get_addr_string(host->addr, buf, len),
+			version);
+	else
+		debug(logopt, "called with host %s version 0x%x",
+			host->name, version);
+
 	memset(&pm_info, 0, sizeof(struct conn_info));
 	memset(&rpc_info, 0, sizeof(struct conn_info));
 	memset(&parms, 0, sizeof(struct pmap));
 
 	if (host->proximity == PROXIMITY_NET)
 		timeout = RPC_TIMEOUT * 2;
-	else if (host->proximity == PROXIMITY_NET)
+	else if (host->proximity == PROXIMITY_OTHER)
 		timeout = RPC_TIMEOUT * 8;
 
 	rpc_info.host = host->name;
+	rpc_info.addr = host->addr;
+	rpc_info.addr_len = host->addr_len;
 	rpc_info.program = NFS_PROGRAM;
 	rpc_info.timeout.tv_sec = timeout;
 	rpc_info.close_option = RPC_CLOSE_DEFAULT;
@@ -652,7 +829,8 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 			return 0;
 	} else {
 		int ret = rpc_portmap_getclient(&pm_info,
-				 host->name, proto, RPC_CLOSE_DEFAULT);
+				host->name, host->addr, host->addr_len,
+				proto, RPC_CLOSE_DEFAULT);
 		if (!ret)
 			return 0;
 
@@ -671,11 +849,14 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 		status = rpc_ping_proto(&rpc_info);
 		gettimeofday(&end, &tz);
 		if (status) {
-			if (random_selection)
+			if (random_selection) {
 				/* Random value between 0 and 1 */
 				taken = ((float) random())/((float) RAND_MAX+1);
-			else
+				debug(logopt, "random selection time %f", taken);
+			} else {
 				taken = elapsed(start, end);
+				debug(logopt, "rpc ping time %f", taken);
+			}
 		}
 	}
 done:
@@ -689,12 +870,17 @@ done:
 
 	if (status) {
 		/* Response time to 7 significant places as integral type. */
-		host->cost = (unsigned long) (taken * 1000000);
+		if (use_weight_only)
+			host->cost = 1;
+		else
+			host->cost = (unsigned long) (taken * 1000000);
 
 		/* Allow for user bias */
 		if (host->weight)
 			host->cost *= (host->weight + 1);
 
+		debug(logopt, "cost %ld weight %d", host->cost, host->weight);
+
 		return 1;
 	}
 
@@ -702,8 +888,7 @@ done:
 }
 
 int prune_host_list(unsigned logopt, struct host **list,
-		    unsigned int vers, const char *options,
-		    unsigned int random_selection)
+		    unsigned int vers, const char *options)
 {
 	struct host *this, *last, *first;
 	struct host *new = NULL;
@@ -724,6 +909,7 @@ int prune_host_list(unsigned logopt, struct host **list,
 	this = first;
 	while (this && this->proximity == PROXIMITY_LOCAL)
 		this = this->next;
+	first = this;
 
 	/*
 	 * Check for either a list containing only proximity local hosts
@@ -735,8 +921,6 @@ int prune_host_list(unsigned logopt, struct host **list,
 		return 1;
 
 	proximity = this->proximity;
-	first = this;
-	this = first;
 	while (this) {
 		struct host *next = this->next;
 
@@ -744,8 +928,7 @@ int prune_host_list(unsigned logopt, struct host **list,
 			break;
 
 		if (this->name) {
-			status = get_vers_and_cost(logopt, this, vers,
-						   options, random_selection);
+			status = get_vers_and_cost(logopt, this, vers, options);
 			if (!status) {
 				if (this == first) {
 					first = next;
@@ -758,6 +941,15 @@ int prune_host_list(unsigned logopt, struct host **list,
 		this = next;
 	}
 
+	/*
+	 * The list of hosts that aren't proximity local may now
+	 * be empty if we haven't been able probe any so we need
+	 * to check again for a list containing only proximity
+	 * local hosts.
+	 */
+	if (!first)
+		return 1;
+
 	last = this;
 
 	/* Select NFS version of highest number of closest servers */
@@ -792,18 +984,31 @@ int prune_host_list(unsigned logopt, struct host **list,
 	max_udp_count = mmax(v4_udp_count, v3_udp_count, v2_udp_count);
 	max_count = max(max_tcp_count, max_udp_count);
 
-	if (max_count == v4_tcp_count)
+	if (max_count == v4_tcp_count) {
 		selected_version = NFS4_TCP_SUPPORTED;
-	else if (max_count == v3_tcp_count)
+		debug(logopt,
+		      "selected subset of hosts that support NFS4 over TCP");
+	} else if (max_count == v3_tcp_count) {
 		selected_version = NFS3_TCP_SUPPORTED;
-	else if (max_count == v2_tcp_count)
+		debug(logopt,
+		      "selected subset of hosts that support NFS3 over TCP");
+	} else if (max_count == v2_tcp_count) {
 		selected_version = NFS2_TCP_SUPPORTED;
-	else if (max_count == v4_udp_count)
+		debug(logopt,
+		      "selected subset of hosts that support NFS2 over TCP");
+	} else if (max_count == v4_udp_count) {
 		selected_version = NFS4_UDP_SUPPORTED;
-	else if (max_count == v3_udp_count)
+		debug(logopt,
+		      "selected subset of hosts that support NFS4 over UDP");
+	} else if (max_count == v3_udp_count) {
 		selected_version = NFS3_UDP_SUPPORTED;
-	else if (max_count == v2_udp_count)
+		debug(logopt,
+		      "selected subset of hosts that support NFS3 over UDP");
+	} else if (max_count == v2_udp_count) {
 		selected_version = NFS2_UDP_SUPPORTED;
+		debug(logopt,
+		      "selected subset of hosts that support NFS2 over UDP");
+	}
 
 	/* Add local and hosts with selected version to new list */
 	this = *list;
@@ -832,8 +1037,7 @@ int prune_host_list(unsigned logopt, struct host **list,
 			add_host(&new, this);
 		} else {
 			status = get_supported_ver_and_cost(logopt, this,
-						selected_version, options,
-						random_selection);
+						selected_version, options);
 			if (status) {
 				this->version = selected_version;
 				remove_host(list, this);
@@ -849,62 +1053,127 @@ int prune_host_list(unsigned logopt, struct host **list,
 	return 1;
 }
 
-static int add_host_addrs(struct host **list, const char *host, unsigned int weight)
+static int add_new_host(struct host **list,
+			const char *host, unsigned int weight,
+			struct addrinfo *host_addr,
+			unsigned int rr, unsigned int options)
 {
-	struct hostent he;
-	struct hostent *phe = &he;
-	struct hostent *result;
-	struct sockaddr_in saddr;
-	char buf[MAX_IFC_BUF], **haddr;
-	int ghn_errno, ret;
 	struct host *new;
 	unsigned int prx;
+	int addr_len;
 
-	saddr.sin_family = AF_INET;
-	if (inet_aton(host, &saddr.sin_addr)) {
-		const char *thost = (const char *) &saddr.sin_addr;
-
-		prx = get_proximity(thost, sizeof(saddr.sin_addr));
-		if (prx == PROXIMITY_ERROR)
-			return 0;
-
-		if (!(new = new_host(host, thost, prx, weight)))
-			return 0;
-
-		if (!add_host(list, new))
-			free_host(new);
-
-		return 1;
+	/*
+	 * If we are using random selection we pretend all hosts are at
+	 * the same proximity so hosts further away don't get excluded.
+	 * We can't use PROXIMITY_LOCAL or we won't perform an RPC ping
+	 * to remove hosts that may be down.
+	 */
+	if (options & MOUNT_FLAG_RANDOM_SELECT)
+		prx = PROXIMITY_SUBNET;
+	else {
+		prx = get_proximity(host_addr->ai_addr);
+		/*
+		 * If we want the weight to be the determining factor
+		 * when selecting a host then all hosts must have the
+		 * same proximity. However, if this is the local machine
+		 * it should always be used since it is certainly available.
+		 */
+		if (prx != PROXIMITY_LOCAL &&
+		   (options & MOUNT_FLAG_USE_WEIGHT_ONLY))
+			prx = PROXIMITY_SUBNET;
 	}
 
-	memset(buf, 0, MAX_IFC_BUF);
-	memset(&he, 0, sizeof(struct hostent));
+	/*
+	 * If we tried to add an IPv6 address and we don't have IPv6
+	 * support return success in the hope of getting an IPv4
+	 * address later.
+	 */
+	if (prx == PROXIMITY_UNSUPPORTED)
+		return 1;
+	if (prx == PROXIMITY_ERROR)
+		return 0;
 
-	ret = gethostbyname_r(host, phe,
-			buf, MAX_IFC_BUF, &result, &ghn_errno);
-	if (ret || !result) {
-		if (ghn_errno == -1)
-			logmsg("host %s: lookup failure %d", host, errno);
-		else
-			logmsg("host %s: lookup failure %d", host, ghn_errno);
+	addr_len = sizeof(struct sockaddr);
+	new = new_host(host, host_addr->ai_addr, addr_len, prx, weight, options);
+	if (!new)
+		return 0;
+
+	if (!add_host(list, new)) {
+		free_host(new);
+		return 0;
+	}
+	new->rr = rr;
+
+	return 1;
+}
+
+static int add_host_addrs(struct host **list, const char *host,
+			  unsigned int weight, unsigned int options)
+{
+	struct addrinfo hints, *ni, *this;
+	char *name = strdup(host);
+	int len;
+	char buf[MAX_ERR_BUF];
+	int rr = 0;
+	int ret;
+
+	if (!name) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		error(LOGOPT_ANY, "strdup: %s", estr);
+		error(LOGOPT_ANY, "failed to add host %s", host);
+		return 0;
+	}
+	len = strlen(name);
+
+	if (name[0] == '[' && name[--len] == ']')
+		name[len] = '\0';
+
+	memset(&hints, 0, sizeof(hints));
+	hints.ai_flags = AI_NUMERICHOST;
+	hints.ai_family = AF_UNSPEC;
+	hints.ai_socktype = SOCK_DGRAM;
+
+	ret = getaddrinfo(name + 1, NULL, &hints, &ni);
+	if (ret)
+		goto try_name;
+
+	this = ni;
+	while (this) {
+		ret = add_new_host(list, host, weight, this, 0, options);
+		if (!ret)
+			break;
+		this = this->ai_next;
+	}
+	freeaddrinfo(ni);
+	goto done;
+
+try_name:
+	memset(&hints, 0, sizeof(hints));
+	hints.ai_flags = AI_ADDRCONFIG;
+	hints.ai_family = AF_UNSPEC;
+	hints.ai_socktype = SOCK_DGRAM;
+
+	ret = getaddrinfo(name + 1, NULL, &hints, &ni);
+	if (ret) {
+		error(LOGOPT_ANY, "hostname lookup failed: %s",
+		      gai_strerror(ret));
+		free(name);
 		return 0;
 	}
 
-	for (haddr = phe->h_addr_list; *haddr; haddr++) {
-		prx = get_proximity(*haddr, phe->h_length);
-		if (prx == PROXIMITY_ERROR)
-			return 0;
-
-		if (!(new = new_host(host, *haddr, prx, weight)))
-			return 0;
-
-		if (!add_host(list, new)) {
-			free_host(new);
-			continue;
-		}
+	this = ni;
+	if (this->ai_next)
+		rr++;
+	while (this) {
+		ret = add_new_host(list, host, weight, this, rr, options);
+		if (!ret)
+			break;
+		this = this->ai_next;
 	}
-
-	return 1;
+	freeaddrinfo(ni);
+done:
+	free(name);
+	return ret;
 }
 
 static int add_path(struct host *hosts, const char *path, int len)
@@ -953,7 +1222,8 @@ static int add_local_path(struct host **hosts, const char *path)
 	new->path = tmp;
 	new->proximity = PROXIMITY_LOCAL;
 	new->version = NFS_VERS_MASK;
-	new->name = new->addr = NULL;
+	new->name = NULL;
+	new->addr = NULL;
 	new->weight = new->cost = 0;
 
 	add_host(hosts, new);
@@ -961,7 +1231,30 @@ static int add_local_path(struct host **hosts, const char *path)
 	return 1;
 }
 
-int parse_location(unsigned logopt, struct host **hosts, const char *list)
+static char *seek_delim(const char *s)
+{
+	const char *p = s;
+	char *delim;
+
+	delim = strpbrk(p, "(, \t:");
+	if (delim && *delim != ':' && (delim == s || *(delim - 1) != '\\'))
+		return delim;
+
+	while (*p) {
+		if (*p != ':') {
+			p++;
+			continue;
+		}
+		if (!strncmp(p, ":/", 2))
+			return (char *) p;
+		p++;
+	}
+
+	return NULL;
+}
+
+int parse_location(unsigned logopt, struct host **hosts,
+		   const char *list, unsigned int options)
 {
 	char *str, *p, *delim;
 	unsigned int empty = 1;
@@ -980,7 +1273,7 @@ int parse_location(unsigned logopt, struct host **hosts, const char *list)
 		int weight = 0;
 
 		p += strspn(p, " \t,");
-		delim = strpbrk(p, "(, \t:");
+		delim = seek_delim(p);
 
 		if (delim) {
 			if (*delim == '(') {
@@ -993,6 +1286,12 @@ int parse_location(unsigned logopt, struct host **hosts, const char *list)
 					*delim = '\0';
 					weight = atoi(w);
 				}
+				else {
+					/* syntax error - Mismatched brackets */
+					free_host_list(hosts);
+					free(str);
+					return 0;
+				}
 				delim++;
 			}
 
@@ -1004,7 +1303,7 @@ int parse_location(unsigned logopt, struct host **hosts, const char *list)
 
 				/* Oh boy - might have spaces in the path */
 				next = path;
-				while (*next && *next != ':')
+				while (*next && strncmp(next, ":/", 2))
 					next++;
 
 				/* No spaces in host names at least */
@@ -1016,7 +1315,7 @@ int parse_location(unsigned logopt, struct host **hosts, const char *list)
 				}
 
 				if (p != delim) {
-					if (!add_host_addrs(hosts, p, weight)) {
+					if (!add_host_addrs(hosts, p, weight, options)) {
 						if (empty) {
 							p = next;
 							continue;
@@ -1038,7 +1337,7 @@ int parse_location(unsigned logopt, struct host **hosts, const char *list)
 				*delim = '\0';
 				next = delim + 1;
 
-				if (!add_host_addrs(hosts, p, weight)) {
+				if (!add_host_addrs(hosts, p, weight, options)) {
 					p = next;
 					continue;
 				}
